/********************************************************************************
 * CALZOLED permette da agire sulla GPIO tramite filesystema /proc
 * Version:  0.1
 * Author:   Calzo
 * Arch:     AXIS ETRAX 100 LX Risc 32 bit processor
 * License:  DUAL BSD/GPL
 * ------------------------------------------------------------------------------
 * Questo modulo non ha bisogno di un file di interfaccia in /dev
 * ------------------------------------------------------------------------------
 * NOTE: 
 * - l'abbreviazione "cl" indica calzoled
 * - "La completezza  nemica della chiarezza" per questo alcune funzioni sono
 *   implementate volutamente in modo incompleto (quindi con alta probabilit di
 *   inserire qualche baco); si far comunque in modo di indicare ogni mancanza
 * ******************************************************************************/

#include <linux/module.h>
#include <linux/proc_fs.h> //provide proc interface
#include <asm/uaccess.h>   //provide copy_from_user, copy_to_user, access_ok, ecc
#include <asm/semaphore.h> //provide semaphore structure
#include <asm/io.h>        //provide access to GPIO port and other peripherals on ETRAX100LX

MODULE_LICENSE("DUAL GPL/BSD");
MODULE_VERSION("0.1");

/** Parametri del modulo 
 * */
static unsigned int setdataPA = 0x00;
module_param(setdataPA, uint, S_IRUGO|S_IWUSR); // see LDD3 cap 2, pag 36
MODULE_PARM_DESC(setdataPA, "Set PA port");

struct semaphore sem; ///mutual exclusion semaphore - serve per evitare le cos dette "Concurrency and Race conditions" (see LDD3 cap5)

//Strutture relative a /proc filesystem
struct proc_dir_entry *proc_cldir = NULL; //main directory
struct proc_dir_entry *proc_clpa  = NULL; //file relativo alla porta PA

/** Strutture relative al filesystem proc
 * */
#define PROC_CL_DIR "calzoled"
#define PROC_CL_PA  "pa"

/*******************************************
 * Funzioni di gestione del filesystem proc
 * *****************************************/

/** _write_pa scrive lo stato dei LED della porta PA. Operando sui LED installati
 * di default sulla FOX8+32  sufficiente operare sui primi 4 bit. La funzione
 * quindi prende in esame solo il primo carattere di un buffer ed esamina il
 * valore interpretandolo come un valore esadecimale. Per esempio:
 *   # echo F > /proc/calzoled/pa
 * equivalr a scrivere *R_PORT_PA_DATA = 15 spegnendo tutti i led.
 * */
int cl_write_pa(struct file *file, const char __user *buffer,
		unsigned long count, void *data)
{
   unsigned char buf=0;
   int new_value=0; //se il nuovo valore  negativo c' un errore
   
   if(!count)      //se il contatore  0...
     return count; //...ritorna 0 e non fare nulla
   
   if(!buffer)  //se il buffer non esiste...
     return 0;  //...per ora ritorno 0, ma sarebbe meglio restituire un errore oppure count
   
   if(down_interruptible(&sem)) //evitaro le concurrency conditions
     return -ERESTARTSYS;
   
   // se sono qui, procedo con la copia del buffer passato dall'user-space 
   // per un massimo della linghezza del buffer di destinazione
   copy_from_user((void*)&buf, buffer, 1); //copia il buffer da use a kernel space
   
   //converto il carattere ESADECIMALE in un numero
   if ( buf>='0' && buf<='9')
     new_value += ((int)buf-'0');
   else if ( buf>='A' && buf<='F')
     new_value += ((int)buf-'A'+10);
   else if ( buf>='a' && buf<='f')
     new_value += ((int)buf-'a'+10);
   else 
	new_value = -1;
   
   // Setto i dati della porta A se non ci sono errori
   if(new_value<0)
     printk(KERN_ALERT "Warning: %c is not HEX format.\n", buf);
   else
     *R_PORT_PA_DATA = new_value;
   
   up(&sem); //rilascio il semaforo
   return count; //ritornando subito count, si fa tutto in una sessione
}


/** _read_pa legge lo stato della porta A restituendo due caratteri che 
 * rappresentano un valore esadecimale pi il carattere di ritorno
 * */
int cl_read_pa(char *buf, char **start, off_t offset,
	       int count, int *eof, void *data)
{
   char str[3]={0,0,0};
   int  len;
   
   if(down_interruptible(&sem))
     return -ERESTARTSYS;
   
   len = sprintf(str, "%2X", *R_PORT_PA_DATA);
   copy_to_user((char __user*)buf, str, len);
   
   up(&sem);
   return 0;
}



/****************************************
 * CLEANUP & INITIALIZATION functions
 ****************************************/
/** Funzione di scaricamento del modulo
 *  * */
void cl_cleanup_module(void)
{
   
   // ELIMINAZIONE delle entit in proc
   if(proc_clpa)
     remove_proc_entry(PROC_CL_PA, proc_cldir); //file e relativa parent directory
   if(proc_cldir)
     remove_proc_entry(PROC_CL_DIR, NULL);
   
   printk(KERN_ALERT "calzoled scaricato\n");
}

/** Funzione di inizializzazione del modulo
 * */
int cl_init_module(void)
{
   //Inizializzazione dei semafori
   init_MUTEX(&sem);
   
   //Creazione della directory virtuale
   proc_cldir = proc_mkdir(PROC_CL_DIR, NULL); //NULL=creo la dir in /proc chiamata PROC_CL_DIR
   if(!proc_cldir)
     {
	printk(KERN_ALERT "Unable to create proc_cldir\n");
	return -ENOMEM;
     } else
     {
	proc_clpa = create_proc_read_entry(PROC_CL_PA, //nome del file
					   0, //protection mask: default=0
					   proc_cldir, //parent dir
					   cl_read_pa, //funzione di lettura del file
					   NULL);
	if(!proc_clpa)
	  {
	     printk(KERN_ALERT "Unable to create proc_clpa\n");
	     cl_cleanup_module();
	     return -ENOMEM;
	  } else
	  {
	      proc_clpa->write_proc = cl_write_pa;
	  }
     }
   
   printk(KERN_ALERT "calzoled caricato.\n PA data:      %x\n PA direction: %x\n",
	  *R_PORT_PA_DATA, *R_PORT_PA_DIR);
   return 0;
}

/***************************************
 * INIT & EXIT
 ***************************************/
module_init(cl_init_module);
module_exit(cl_cleanup_module);
